//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_ilda.hpp"

//############################################################################
ez_3D_float ez_real_segment::segment_front() const
{
    ez_3D_float front_;
    front_.z = -FLT_MAX;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).z > front_.z)
            front_ = at(i);
    return front_;
}

//############################################################################
ez_3D_float ez_real_segment::segment_back() const
{
    ez_3D_float back_;
    back_.z = FLT_MAX;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).z < back_.z)
            back_ = at(i);
    return back_;
}

//############################################################################
ez_3D_float ez_real_segment::segment_top() const
{
    ez_3D_float top_;
    top_.y = -FLT_MAX;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).y > top_.y)
            top_ = at(i);
    return top_;
}

//############################################################################
ez_3D_float ez_real_segment::segment_bottom() const
{
    ez_3D_float bottom_;
    bottom_.y = FLT_MAX;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).y < bottom_.y)
            bottom_ = at(i);
    return bottom_;
}

//############################################################################
ez_3D_float ez_real_segment::segment_right() const
{
    ez_3D_float right_;
    right_.x = -FLT_MAX;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).x > right_.x)
            right_ = at(i);
    return right_;
}

//############################################################################
ez_3D_float ez_real_segment::segment_left() const
{
    ez_3D_float left_;
    left_.x = FLT_MAX;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).x < left_.x)
            left_ = at(i);
    return left_;
}

//############################################################################
float ez_real_segment::segment_height() const
{
    return (segment_top().y - segment_bottom().y);
}
//############################################################################
float ez_real_segment::segment_width() const
{
    return (segment_right().x - segment_left().x);
}

//############################################################################
float ez_real_segment::segment_depth() const
{
    return (segment_front().z - segment_back().z);
}

//############################################################################
float ez_real_segment::segment_size() const
{
    float greatest = 0;
    if(segment_width () > greatest)    greatest = segment_width ();
    if(segment_height() > greatest)    greatest = segment_height();
    if(segment_depth () > greatest)    greatest = segment_depth ();
    return greatest;
}

//############################################################################
void ez_real_segment::blank_all_points()
{
    for(int i = 1; i < number_of_points(); i++)
        at(i).unblank();
    return;
}

//############################################################################
void ez_real_segment::unblank_all_points()
{
    for(int i = 1; i < number_of_points(); i++)
        at(i).unblank();
    return;
}

//############################################################################
ez_real_segment& ez_real_segment::reverse()
{
    if(number_of_points() > 1)
    {
        int             i;
        ez_real_segment   reversed;
        ez_real_point   point;
        //--------------------------------------------------------------------
        i = number_of_points() - 1;
        //--------------------------------------------------------------------
        point = at(i--);
        point.blank();
        reversed.push_back(point);       
        //--------------------------------------------------------------------
        while(i >= 0)
        {
            point   = at(i    );
            point.r = at(i + 1).r;
            point.g = at(i + 1).g;
            point.b = at(i + 1).b;
            point.c = at(i + 1).c;
            point.k = at(i + 1).k;
            reversed.push_back(point);
            i--;
        }
        //--------------------------------------------------------------------
        *this = reversed;
    }
    return *this;
}

//############################################################################
void ez_real_segment::flip(int plane)
{
    if(number_of_points() > 1)
    {
        int i;
        switch(plane)
        {
           case 0: // X mirror
                   for(i = 0; i < number_of_points(); i++)
                       at(i).x = -at(i).x;
                   break;                                
           //-----------------------------------------------------------------
           case 1: // Y flip
                   for(i = 0; i < number_of_points(); i++)
                       at(i).y = -at(i).y;
                   break;
           //-----------------------------------------------------------------
           case 2: // Z
                   for(i = 0; i < number_of_points(); i++)
                       at(i).z = -at(i).z;
                   break;
           //-----------------------------------------------------------------
           case 3: // X, Y
                   for(i = 0; i < number_of_points(); i++)
                   {
                       at(i).x = -at(i).x;
                       at(i).y = -at(i).y;
                   }
                   break;                                
           //-----------------------------------------------------------------
           case 4: // X, Y, Z
                   for(i = 0; i < number_of_points(); i++)
                       at(i) = -at(i);
                   break;                                
           //-----------------------------------------------------------------
        }
    }
    return;
}

//############################################################################
void ez_real_segment::quarter_turn(int plane, int turns)
{
    if(number_of_points() > 1)
    {
        int   i, j;
        float temp;
        switch(plane)
        {
           case 0: // X Y
                   for(i = 0; i < number_of_points(); i++)
                       for(j = 0; j < turns; j++)
                       {
                           temp = -at(i).x;
                           at(i).x = at(i).y;
                           at(i).y = temp;
                       }
                   break;                                
           //-----------------------------------------------------------------
           case 1: // Z Y
                   for(i = 0; i < number_of_points(); i++)
                       for(j = 0; j < turns; j++)
                       {
                           temp = -at(i).z;
                           at(i).z = at(i).y;
                           at(i).y = temp;
                       }
                   break;
           //-----------------------------------------------------------------
           case 2: // X Z
                   for(i = 0; i < number_of_points(); i++)
                       for(j = 0; j < turns; j++)
                       {
                           temp = -at(i).x;
                           at(i).x = at(i).z;
                           at(i).z = temp;
                       }
                   break;
           //-----------------------------------------------------------------
        }
    }
    return;
}

//############################################################################
void ez_real_segment::flatten_z()
{
    if(number_of_points() > 1)
    {
        for(int i = 0; i < number_of_points(); i++)
            at(i).z = 0;
    }
    return;
}

//############################################################################
int ez_real_segment::rotate_around_origin(ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        for(int i = 0; i < number_of_points(); i++)
            at(i) = ez_real_point( rotate_point(at(i), a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_real_segment::rotate_around_point(ez_3D_float p, ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        for(int i = 0; i < number_of_points(); i++)
            at(i) = ez_real_point( rotate_point_around_point(at(i), p, a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
bool ez_real_segment::find_rgb_in_palette(const ez_palette& palette)
{
    int           i,
                  j;
    vector<bool>  match(number_of_points(), false);
    bool          all_colors_found = true;
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_points(); i++)
        for(j = 0; j < palette.number_of_colors(); j++)
            if((ez_color)at(i) == palette.at(j))
            {
                at(i).c = (u_char)j;
                match[i] = true;
            }
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_points(); i++)
        all_colors_found &= match[i];
    //------------------------------------------------------------------------
    return all_colors_found;
}

//############################################################################
void ez_real_segment::set_rgb_from_palette()
{
    at(0).r = 
    at(0).g = 
    at(0).b = 0x00;
    //------------------------------------------------------------------------
    for(int i = 1; i < number_of_points(); i++)
    {
        at(i).r = ez_ilda_pointer->palette_picker(palette_index)[at(i).c].r;
        at(i).g = ez_ilda_pointer->palette_picker(palette_index)[at(i).c].g;
        at(i).b = ez_ilda_pointer->palette_picker(palette_index)[at(i).c].b;
    }
    return;
}

//############################################################################
void ez_real_segment::set_palette_to_332()
{
    at(0).c = 0x00;
    for(int i = 1; i < number_of_points(); i++)
        at(i).c =    (at(i).r & 0xe0)
                  | ((at(i).g & 0xe0) >> 3)
                  | ((at(i).b & 0xc0) >> 6);
    // does NOT set palette_index to EZ_EZFB_332
    return;
}

//############################################################################
void ez_real_segment::sync_rgb_and_palette()
{
    if(number_of_points() > 1)
    {
        if(palette_index != EZ_TRUE_COLOR)
            set_rgb_from_palette();
        //--------------------------------------------------------------------
        else
        {
            int          i                  ,
                         j                  ;
            char         palette_name[8]    ;
            ez_palette   palette            ;
            static int   palette_number = 0 ;
            //----------------------------------------------------------------
            sprintf(palette_name, "%07d", palette_number);
            palette.name = string("P") + palette_name;
            palette.push_back((ez_color)at(1));
            //----------------------------------------------------------------
            for(i = 1; i < number_of_points(); i++)
            {
                for(j = 0; j < palette.number_of_colors(); j++)
                    if(palette[j] == (ez_color)at(i))
                        break;
                if(j == palette.number_of_colors())
                    palette.push_back((ez_color)at(i));

                if(palette.number_of_colors() > EZ_PALETTE_MAX)
                    break;
            }
            //----------------------------------------------------------------
            if(palette.number_of_colors() <= EZ_PALETTE_MAX)
            {
                palette.reorder();
                palette.find_factors();
                find_rgb_in_palette(palette);
                ez_ilda_pointer->push_back_palette(palette);
                palette_index = ez_ilda_pointer->number_of_palettes() - 1;
                palette_number++;
            }
            //----------------------------------------------------------------
            else // (palette.number_of_colors() > EZ_PALETTE_MAX)
                set_palette_to_332();
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ez_real_segment::bit_reduce_to_palette()
{
    if(palette_index == EZ_TRUE_COLOR)
    {
        set_palette_to_332();
        palette_index = EZ_EZFB_332;
        set_rgb_from_palette();
    }
    return;
}

//############################################################################
int ez_real_segment::move(ez_3D_float d)
{
    if(number_of_points() > 1)
    {
        ez_real_segment segment(*this);
        int        i,
                   out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            out_of_bounds = ez_ilda_bounds_check(at(i).position() + d, EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
            at(i) = ez_real_point( at(i).position() + d,
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_real_segment::scale(ez_3D_float s)
{
    ez_3D_float center = find_center();
    return scale_around_point(center, s);
}

//############################################################################
int ez_real_segment::scale_around_point(ez_3D_float p, ez_3D_float s)
{
    if(number_of_points() > 1)
    {
        ez_real_segment segment(*this);
        int        i,
                   out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            out_of_bounds = ez_ilda_bounds_check(((at(i).position() - p) * s) + p, EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
            at(i) = ez_real_point( ((at(i).position() - p) * s) + p,
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_real_segment::scale_around_origin(ez_3D_float s)
{
    if(number_of_points() > 1)
    {
        ez_real_segment segment(*this);
        int        i,
                   segment_index,
                   out_of_bounds = EZ_IN_BOUNDS;

        for(segment_index = 0; segment_index < segment.number_of_segments(); segment_index++)
            if((out_of_bounds = segment.scale_segment_around_origin(segment_index, s)))
                return out_of_bounds;

        for(i = 0; i < number_of_points(); i++)
            at(i) *= s;
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
ez_3D_float ez_real_segment::find_center() const
{
    ez_3D_float  center;
    if(number_of_points() > 1)
    {
        int     i;

        float   max_x = -FLT_MAX,
                min_x =  FLT_MAX,
                max_y = -FLT_MAX,
                min_y =  FLT_MAX,
                max_z = -FLT_MAX,
                min_z =  FLT_MAX;

        for(i = 0; i < number_of_points(); i++)
        {
            if(at(i).x > max_x)    max_x = at(i).x;
            if(at(i).x < min_x)    min_x = at(i).x;
            if(at(i).y > max_y)    max_y = at(i).y;
            if(at(i).y < min_y)    min_y = at(i).y;
            if(at(i).z > max_z)    max_z = at(i).z;
            if(at(i).z < min_z)    min_z = at(i).z;
        }
        center.x = (max_x - ((max_x - min_x) / 2.0));
        center.y = (max_y - ((max_y - min_y) / 2.0));
        center.z = (max_z - ((max_z - min_z) / 2.0));
    }
    return center;
}

//############################################################################
int ez_real_segment::number_of_segments() const // a segment is a series of lit points
{
    int   i,
          segment_count = 0;
    if(number_of_points() > 1)
    {
        for(i = 1; i < number_of_points(); i++)
        {
            if(at(i).is_lit())
            {
                while(at(i).is_lit() && i < (number_of_points() - 1))
                    i++;
                segment_count++;
            }
        }
    }
    return segment_count;
}

//############################################################################
bool ez_real_segment::find_segment(int segment_index, int& start, int& end) const
{   // the first segment is number zero!
    if(number_of_points() > 1)
    {
        int  i,
             segment_count = -1;
        //--------------------------------------------------------------------
        for(i = 1; i < number_of_points(); i++)
        {
            if(at(i).is_lit())
            {
                start = i - 1;
                while(at(i).is_lit() && i < (number_of_points() - 1))
                    i++;
                //------------------------------------------------------------
                end = i - 1;
                if(    i == (number_of_points() - 1)
                    && at(i).is_lit()
                  )
                {
                    end = i;
                }
                //------------------------------------------------------------
                segment_count++;
                if(segment_count == segment_index)
                    return true;
            }
        } // segment index either negative or out of range
        //--------------------------------------------------------------------
        if(segment_index > -1)
        {
            for(i = 1; i < number_of_points(); i++)
            {
                if(at(i).is_lit())
                {
                    start = i - 1;
                    while(at(i).is_lit() && i < (number_of_points() - 1))
                        i++;
                    end = i - 1;
                    if(    i == (number_of_points() - 1)
                        && at(i).is_lit()
                      )
                        end = i;
                    return false; // and set start, end to first segment
                }
            }
        }
    }
    //------------------------------------------------------------------------
    return false;
}

//############################################################################
ez_real_segment ez_real_segment::find_segment(int segment_index) const
{
    ez_real_segment segment;
    if(number_of_points() > 1)
    {
        int start, end;
        if(find_segment(segment_index, start, end))
        {
            segment.clear();
            for(int i = start; i <= end; i++)
                segment.push_back(at(i));           
        }
    }
    return segment;
}

//############################################################################
ez_3D_float ez_real_segment::find_segment_center(int segment_index) const
{
    return (find_segment(segment_index)).find_center();
}

//############################################################################
int ez_real_segment::move_segment(int segment_index, ez_3D_float f)
{
    if(number_of_points() > 1)
    {
        int  i            ,
             out_of_bounds,
             start        ,
             end          ;

        find_segment(segment_index, start, end);

        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(f + at(i), EZ_ILDA_CUBE);
            if(out_of_bounds != EZ_IN_BOUNDS)
                return out_of_bounds;
        }
        for(i = start; i <= end; i++)
            at(i) += f;
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_real_segment::rotate_segment(int segment_index, ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        int          i            ,
                     out_of_bounds,
                     start        ,
                     end          ;

        ez_3D_float  center       ;

        find_segment(segment_index, start, end);
        center = find_segment_center(segment_index);

        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point(at(i), center, a), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = start; i <= end; i++)
            at(i) = ez_real_point( rotate_point_around_point(at(i), center, a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_real_segment::rotate_segment_around_origin(int segment_index, ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        int  i            ,
             out_of_bounds,
             start        ,
             end          ;
        find_segment(segment_index, start, end);
        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point(at(i), a), EZ_ILDA_SPHERE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = start; i <= end; i++)
            at(i) = ez_real_point( rotate_point(at(i), a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_real_segment::scale_segment(int segment_index, ez_3D_float m)
{
    if(number_of_points() > 1)
    {
        int          i,
                     out_of_bounds,
                     start,
                     end;
        ez_3D_float  center;
        //--------------------------------------------------------------------
        find_segment(segment_index, start, end);
        center = find_segment_center(segment_index);
        //--------------------------------------------------------------------
        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(scale_point_around_point(at(i), center, m), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        //--------------------------------------------------------------------
        for(i = start; i <= end; i++)
            at(i) = ez_real_point( scale_point_around_point(at(i), center, m),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c);
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_real_segment::scale_segment_around_origin(int segment_index, ez_3D_float m)
{
    if(number_of_points() > 1)
    {
        int  i            ,
             out_of_bounds,
             start        ,
             end          ;
        //--------------------------------------------------------------------
        find_segment(segment_index, start, end);
        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(m * at(i), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        //--------------------------------------------------------------------
        for(i = start; i <= end; i++)
            at(i) = ez_real_point( m * at(i),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_real_segment::from_ifstream_dxf(ifstream& in)
{
    int             group_code                ,
                    arc_points    = 0         ,
                    first_segment_point_index ,
                    group_70_flag = 0         ,
                    group_71_flag = 0         ,
                    group_72_flag = 0         ,
                    group_73_flag = 0         ;
       
    string          entity_string             ,
                    font                      ,
                    text                      ;
    
    u_char          color                     ,
                    prev_color                ;

    float           float_value               ,
                    radius                    ,
                    arc_start                 ,
                    arc_end                   ,
                    arc_angle                 ,
                    arc_step                  ;

    ez_3D_float     float_3D_1                ,
                    float_3D_2                ,
                    float_3D_3                ,
                    float_3D_4                ;

    ez_real_segment _real_points              ;

    vector<float>   vertex_x                  ,
                    vertex_y                  ;

    bool            closed_polyline           ;
    //------------------------------------------------------------------------
    while(!in.eof() && entity_string != "ENTITIES") // ignore everthing up to ENTITIES
        in >> entity_string;
      //------------------------------------------------------------------------
    if(!in.eof())
    {
        while(!in.eof())
        {
            color                      = ez_ilda_pointer->palette_picker(EZ_DXF).white;
            prev_color                 = ez_ilda_pointer->palette_picker(EZ_DXF).white;
            first_segment_point_index  = -1,
            closed_polyline            = false;
            group_70_flag              = 0;
            radius                     = 0;
            arc_points                 = 0;
            arc_start                  = 0;
            arc_end                    = 360;
            group_code                 = -1;
            vertex_x.clear();
            vertex_y.clear();
            //----------------------------------------------------------------
            if(entity_string == "POINT")
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                    }
                }
                push_back(ez_real_point(float_3D_1, color).blank());
                push_back(ez_real_point(float_3D_1, color).unblank());
            }
            else if(entity_string == "LINE")
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 11: // x2
                                    float_3D_2.x = float_value;
                                    break;
                        case 21: // y2
                                    float_3D_2.y = float_value;
                                    break;
                        case 31: // z2
                                    float_3D_2.z = float_value;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                    }
                }
                push_back(ez_real_point(float_3D_1, color).blank());
                push_back(ez_real_point(float_3D_2, color).unblank());
            }
            //------------------------------------------------------------
            else if(    entity_string == "CIRCLE"
                     || entity_string == "ARC"
                   )
            {
                int a;
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 40: // radius
                                    radius = float_value;
                                    break;
                        case 50: // start angle
                                    arc_start = float_value;
                                    break;
                        case 51: // end angle
                                    arc_end = float_value;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                    }
                }

                if(arc_start >= arc_end)
                    arc_end += 360.0;
                arc_angle  = arc_end - arc_start;
                arc_points = (int)ceil(arc_angle / ez_ilda_pointer->rendered_arc_angle);
                arc_step   = arc_angle / arc_points;
                arc_start *= one_degree;
                arc_end   *= one_degree;
                arc_step  *= one_degree;

                float_3D_2.x = radius * cos(arc_start);
                float_3D_2.y = radius * sin(arc_start);
                float_3D_2.z = float_3D_1.z;

                push_back(ez_real_point(float_3D_1 + float_3D_2, color).blank());

                for(a = 1; a <= arc_points; a++)
                {
                    float_3D_2.x = radius * cos(a * arc_step + arc_start);
                    float_3D_2.y = radius * sin(a * arc_step + arc_start);
                    float_3D_2.z = float_3D_1.z;

                    push_back(ez_real_point(float_3D_1 + float_3D_2, color).unblank());
                }
            }
            //------------------------------------------------------------
            else if(entity_string == "ELLIPSE")
            {
                int a,
                    point_index;
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 11: // x2
                                    float_3D_2.x = float_value;
                                    break;
                        case 21: // y2
                                    float_3D_2.y = float_value;
                                    break;
                        case 31: // z2
                                    float_3D_2.z = float_value;
                                    break;
                        case 40: // ratio between major and minor axis
                                    radius = float_value;
                                    break;
                        case 41: // start angle
                                    arc_start = float_value;
                                    break;
                        case 42: // end angle
                                    arc_end = float_value;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                    }
                }
                if(arc_start >= arc_end) arc_end += two_pi;
                arc_angle  = arc_end - arc_start;
                arc_points = (int)ceil(arc_angle / (ez_ilda_pointer->rendered_arc_angle * one_degree));
                arc_step   = arc_angle / arc_points;
                float major_axis = float_3D_2.magnitude(),
                      minor_axis = major_axis * radius;
                float_3D_3   = 0.0;
                float_3D_3.z = -float_3D_2.direction().z;
                _real_points.clear();
                float_3D_4.x = -major_axis * cos(arc_start);
                float_3D_4.y = -minor_axis * sin(arc_start);
                float_3D_4.z = float_3D_1.z;
                _real_points.push_back(ez_real_point(float_3D_4, color).blank());
                for(a = 1; a <= arc_points; a++)
                {
                    float_3D_4.x = -major_axis * cos(a * arc_step + arc_start);
                    float_3D_4.y = -minor_axis * sin(a * arc_step + arc_start);
                    float_3D_4.z = float_3D_1.z;

                    _real_points.push_back(ez_real_point(float_3D_4, color).unblank());
                }
                if(float_3D_3 != 0.0)
                    for(point_index = 0; point_index < (int)_real_points.size(); point_index++)
                        _real_points[point_index] = rotate_point(_real_points[point_index], float_3D_3);
                for(point_index = 0; point_index < (int)_real_points.size(); point_index++)
                    push_back(_real_points[point_index] + float_3D_1);
            }
            //------------------------------------------------------------
            else if(entity_string == "POLYLINE")
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {                    
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 62: // color
                                    color = (u_char)float_value;
                                    prev_color = color;
                                    break;
                        case 70:
                                    if(((int)float_value) & 0x01)
                                        closed_polyline = true;
                                    break;
                    }
                }
                while(entity_string == "VERTEX")
                {
                    while(get_dxf_pair(in, group_code, entity_string))
                    {
                        sscanf(entity_string.c_str(), "%f", &float_value);
                        switch(group_code)
                        {
                            case 10: // x1
                                        float_3D_1.x = float_value;
                                        break;
                            case 20: // y1
                                        float_3D_1.y = float_value;
                                        break;
                            case 30: // z1
                                        float_3D_1.z = float_value;
                                        break;
                            case 62: // color
                                        prev_color = color;
                                        color = (u_char)float_value;
                                        break;
                        }
                    }
                    if(first_segment_point_index == -1)
                    {
                        push_back(ez_real_point(float_3D_1, color).blank());
                        first_segment_point_index = size() - 1;
                    }
                    else
                        push_back(ez_real_point(float_3D_1, prev_color).unblank());
                } // end while(entity_string == "VERTEX")
                if(closed_polyline)
                {
                    push_back(ez_real_point((ez_3D_float)(at(first_segment_point_index)), back().c).unblank());
                }
            }
            //------------------------------------------------------------
            else if(entity_string == "LWPOLYLINE")
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 90: // number of verticies
                                    arc_points = (int)float_value;
                                    break;                        
                        case 70: // is closed polyline
                                    group_70_flag = (int)float_value;
                                    break;                        
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                        case 10: // x
                                    vertex_x.push_back(float_value);
                                    break;
                        case 20: // y
                                    vertex_y.push_back(float_value);
                                    break;
                    }
                }
                if(    (arc_points > 0)
                    && (arc_points == (int)vertex_x.size())
                    && (arc_points == (int)vertex_y.size())
                  )
                {
                    if(color == ez_ilda_pointer->palette_picker(EZ_DXF).black)
                        color = ez_ilda_pointer->palette_picker(EZ_DXF).white;
                    float_3D_1.x = vertex_x[0];
                    float_3D_1.y = vertex_y[0];
                    float_3D_1.z = 0;
                    push_back(ez_real_point(float_3D_1, color).blank());
                    for(int a = 0; a < arc_points; a++)
                    {
                        float_3D_1.x = vertex_x[a];
                        float_3D_1.y = vertex_y[a];
                        push_back(ez_real_point(float_3D_1, color).unblank());
                    }
                    if(group_70_flag & 0x01) // closed polyline
                    {
                        float_3D_1.x = vertex_x[0];
                        float_3D_1.y = vertex_y[0];
                        push_back(ez_real_point(float_3D_1, color).unblank());
                    }
                }
            }
            //------------------------------------------------------------
            else if(entity_string == "TEXT")
            {   
                int               text_index,
                                  point_index;
                ez_ilda_frame_set font_frames;                
                float_3D_1 = 0.0;
                float_3D_3 = 0.0;
                while(get_dxf_pair(in, group_code, entity_string))
                {                    
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case  1: // the text
                                    text = entity_string;
                                    break;
                        case  7: // the font
                                    font = entity_string;
                                    break;
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 40: // text height
                                    radius = float_value;
                                    break;
                        case 50: // flat rotation angle
                                    float_3D_3.z = float_value * one_degree;
                                    break;
                        case 51: // oblique rotation angle
                                    float_3D_3.y = float_value * one_degree;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                        case 71: // flipped or mirrored
                                    group_71_flag = (int)float_value;
                                    break;
                        case 72: // horizontal justification
                                    group_72_flag = (int)float_value;
                                    break;
                        case 73: // vertical justification
                                    group_73_flag = (int)float_value;
                                    break;
                    }
                }
                if(font == "LUCIDA_CONSOLE")
                    font_frames.from_file("lucida.ild");
                else
                    font_frames.from_file("courier_new.ild");
                _real_points.clear();
                for(text_index = 0; text_index < (int)text.size(); text_index++)
                {
                    if(text[text_index] >= '!' && text[text_index] <= '~')
                    {
                        for(point_index = 0; point_index < font_frames[text[text_index] - '!'].quantity; point_index++)
                        {
                            float_3D_2 = font_frames[text[text_index] - '!'].at(point_index);
                            float_3D_2.x = ((float_3D_2.x / (float)24000.0) + (float)(text_index)) * radius;
                            float_3D_2.y = ((float_3D_2.y / (float)24000.0)                      ) * radius;
                            _real_points.push_back(   (font_frames[text[text_index] - '!'].at(point_index).is_blank())
                                                    ? (ez_real_point(float_3D_2, color).blank())
                                                    : (ez_real_point(float_3D_2, color).unblank())
                                                  );
                        }
                    }
                }
                if(group_71_flag & 2) // mirrored
                    _real_points.flip(0);

                if(group_71_flag & 4) // flipped
                    _real_points.flip(1);

                if(float_3D_3 != 0.0)
                    for(point_index = 0; point_index < (int)_real_points.size(); point_index++)
                        _real_points[point_index] = rotate_point(_real_points[point_index], float_3D_3);

                for(point_index = 0; point_index < (int)_real_points.size(); point_index++)
                    push_back(_real_points[point_index] + float_3D_1);
            } // end else if(entity_string == "TEXT")             
            //------------------------------------------------------------
            else if(entity_string == "ENDSEC")
                break;
            //------------------------------------------------------------
            else
                get_dxf_pair(in, group_code, entity_string);
        } // end while(!in.eof())
        set_rgb_from_palette();
    } // end if(!in.eof())
    //------------------------------------------------------------------------
    else
        error = EZ_ILDA_EOF_ERROR;
    //------------------------------------------------------------------------    
    return error;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
